//
// Created by philipp on 01.12.17.
//

#ifndef PROTOCOL_PLATOONCREATEMESSAGE_H
#define PROTOCOL_PLATOONCREATEMESSAGE_H

#include <cstdint>
#include <vector>
#include "Protocol.h"
#include "json.hpp"
#include "NetworkingLib/Message.h"
#include <unordered_set>

namespace platoonProtocol
{

namespace messageTypes
{

constexpr MessageType LV_BROADCAST = 0x00000001;
constexpr MessageType FV_HEARTBEAT = 0x00000002;
constexpr MessageType FV_REQUEST = 0x00000008;
constexpr MessageType LV_ACCEPT = 0x00000010;
constexpr MessageType REJECT = 0x00000020;
constexpr MessageType FV_LEAVE = 0x00000040;

const std::unordered_set<MessageType> standardMessageTypes{
    LV_BROADCAST, FV_HEARTBEAT, FV_REQUEST, LV_ACCEPT, REJECT, FV_LEAVE};

inline bool isStandardMessageType(MessageType messageType)
{
    return messageTypes::standardMessageTypes.find(messageType) != messageTypes::standardMessageTypes.end();
}

template<typename ... MessageTypes>
bool contains(MessageType type, MessageTypes ... types)
{
    std::unordered_set<MessageType> set{types...};
    return set.find(type) != set.end();
}

}

namespace jsonNames
{

constexpr char SRC_VEHICLE[] = "src_vehicle";
constexpr char DST_VEHICLE[] = "dst_vehicle";
constexpr char PLATOON_ID[] = "platoon_id";
constexpr char PLATOON_SPEED[] = "ps";
constexpr char INNER_PLATOON_DISTANCE[] = "ipd";
constexpr char FOLLOWERS[] = "followers";

}

struct PlatoonMessage
{
    using Followers = std::vector<VehicleId>;

    static constexpr std::size_t MAX_SIZE = 256;

    PlatoonMessage() = default;

    PlatoonMessage(
        MessageType messageType,
        VehicleId srcVehicle,
        VehicleId dstVehicle,
        PlatoonId platoonId,
        PlatoonSpeed platoonSpeed,
        InnerPlatoonDistance innerPlatoonDistance,
        const Followers & followers)
        : messageType(messageType)
          , srcVehicle(srcVehicle)
          , dstVehicle(dstVehicle)
          , platoonId(platoonId)
          , platoonSpeed(platoonSpeed)
          , innerPlatoonDistance(innerPlatoonDistance)
          , followers(followers)
    {}

    static PlatoonMessage followerRequest(VehicleId srcVehicle)
    { return PlatoonMessage(messageTypes::FV_REQUEST, srcVehicle, 0, 0, 0, 0, {}); }

    static PlatoonMessage rejectResponse(VehicleId srcVehicle,
                                         VehicleId dstVehicle)
    { return PlatoonMessage(messageTypes::REJECT, srcVehicle, dstVehicle, 0, 0, 0, {}); }

    static PlatoonMessage acceptResponse(VehicleId srcVehicle,
                                         VehicleId dstVehicle,
                                         PlatoonId platoonId,
                                         const Followers & followers)
    { return PlatoonMessage(messageTypes::LV_ACCEPT, srcVehicle, dstVehicle, platoonId, 0, 0, followers); }

    static PlatoonMessage broadcastMessage(VehicleId srcVehicle,
                                           PlatoonId platoonId,
                                           PlatoonSpeed platoonSpeed,
                                           InnerPlatoonDistance innerPlatoonDistance,
                                           const Followers & followers)
    {
        return PlatoonMessage(messageTypes::LV_BROADCAST,
                              srcVehicle,
                              0,
                              platoonId,
                              platoonSpeed,
                              innerPlatoonDistance,
                              followers);
    }

    static PlatoonMessage heartbeatMessage(VehicleId srcVehicle, VehicleId dstVehicle, PlatoonId platoonId)
    { return PlatoonMessage(messageTypes::FV_HEARTBEAT, srcVehicle, dstVehicle, platoonId, 0, 0, {}); }

    static PlatoonMessage leaveMessage(VehicleId srcVehicle, VehicleId dstVehicle, PlatoonId platoonId)
    { return PlatoonMessage(messageTypes::FV_LEAVE, srcVehicle, dstVehicle, platoonId, 0, 0, {}); }

    MessageType messageType{};
    VehicleId srcVehicle{};
    VehicleId dstVehicle{};
    PlatoonId platoonId{};
    PlatoonSpeed platoonSpeed{};
    InnerPlatoonDistance innerPlatoonDistance{};
    Followers followers{};
};

}

namespace networking
{
namespace message
{

template<>
struct Encoder<platoonProtocol::PlatoonMessage>
{
    void operator()(const platoonProtocol::PlatoonMessage & message, std::string & data) const
    {
        using namespace platoonProtocol;

        auto messageType = message.messageType;

        if (!messageTypes::isStandardMessageType(messageType))
            throw 0;    // Throw something to signal a failed encoding.

        // 4 bytes for message type
        data = std::string(4, '\0');

        std::uint8_t numDataBytes[sizeof(MessageType)];
        networking::utils::toBigEndian<4>(numDataBytes, messageType);

        for (std::size_t i = 0; i < sizeof(MessageType); i++)
            data[i] = numDataBytes[i];

        using json = nlohmann::json;
        using namespace jsonNames;
        auto j = json{{SRC_VEHICLE, message.srcVehicle}};

        using namespace messageTypes;
        if (messageTypes::contains(messageType, FV_HEARTBEAT, LV_ACCEPT, REJECT, FV_LEAVE))
            j[DST_VEHICLE] = message.dstVehicle;

        if (messageTypes::contains(messageType, LV_BROADCAST, FV_HEARTBEAT, LV_ACCEPT, FV_LEAVE))
            j[PLATOON_ID] = message.platoonId;

        if (messageTypes::contains(messageType, LV_BROADCAST))
        {
            j[PLATOON_SPEED] = message.platoonSpeed;
            j[INNER_PLATOON_DISTANCE] = message.innerPlatoonDistance;
            j[FOLLOWERS] = message.followers;
        }

        data.append(j.dump());
    }
};

template<>
struct Decoder<platoonProtocol::PlatoonMessage>
{
    void operator()(platoonProtocol::PlatoonMessage & message, const std::string & data) const
    {
        using namespace platoonProtocol;

        if (data.size() < sizeof(MessageType))
            throw 0;    // Throw something to inform that we failed to decode the message.

        using json = nlohmann::json;
        auto j = json::parse(data.begin() + sizeof(MessageType), data.end());

        auto messageType = networking::utils::fromBigEndian<sizeof(MessageType), MessageType>(
            (const std::uint8_t *) data.c_str());

        // We only know message types conforming to the standard protocol specification.
        if (!messageTypes::isStandardMessageType(messageType))
            throw 0;

        message.messageType = messageType;

        using namespace jsonNames;
        message.srcVehicle = j.at(SRC_VEHICLE).get<VehicleId>();

        using namespace messageTypes;
        if (messageTypes::contains(messageType, FV_HEARTBEAT, LV_ACCEPT, REJECT, FV_LEAVE))
            message.dstVehicle = j.at(DST_VEHICLE).get<VehicleId>();

        if (messageTypes::contains(messageType, LV_BROADCAST, FV_HEARTBEAT, LV_ACCEPT, FV_LEAVE))
            message.platoonId = j.at(PLATOON_ID).get<VehicleId>();

        if (messageTypes::contains(messageType, LV_BROADCAST))
        {
            message.platoonSpeed = j.at(PLATOON_SPEED).get<PlatoonSpeed>();
            message.innerPlatoonDistance = j.at(INNER_PLATOON_DISTANCE).get<InnerPlatoonDistance>();
            message.followers = j.at(FOLLOWERS).get<PlatoonMessage::Followers>();
        }
    };
};

}
}

#endif //PROTOCOL_PLATOONCREATEMESSAGE_H
